import json
from typing import Any, Mapping, Optional

from aiohttp import web


class ConfigurationError(Exception):
    invalid_data: Mapping[str, Any]

    def __init__(self, invalid_data: Mapping[str, Any]) -> None:
        super().__init__(invalid_data)
        self.invalid_data = invalid_data


class InvalidAPIHandlerDefinition(Exception):
    pass


class UnknownImageReference(ValueError):
    """
    Represents an error for invalid/unknown image reference.
    The first argument of this exception should be the reference given by the user.
    """

    def __str__(self) -> str:
        return f"Unknown image reference: {self.args[0]}"


class ImageNotAvailable(ValueError):
    """
    Represents an error for unavailability of the image in agents.
    The first argument of this exception should be the reference given by the user.
    """

    def __str__(self) -> str:
        return f"Unavailable image in the agent: {self.args[0]}"


class UnknownImageRegistry(ValueError):
    """
    Represents an error for invalid/unknown image registry.
    The first argument of this exception should be the registry given by the user.
    """

    def __str__(self) -> str:
        return f"Unknown image registry: {self.args[0]}"


class InvalidImageName(ValueError):
    """
    Represents an invalid string for image name.
    """

    def __str__(self) -> str:
        return f"Invalid image name: {self.args[0]}"


class InvalidImageTag(ValueError):
    """
    Represents an invalid string for image tag and full image name.
    Image tag should be a string of form below

    ```
    <version-stringA>-<platform-tag-1A>-<platform-tag-2A>-....
    ```
    """

    def __init__(self, tag: str, full_name: str | None = None) -> None:
        super().__init__(tag, full_name)
        self._tag = tag
        self._full_name = full_name

    def __str__(self) -> str:
        if self._full_name is not None:
            return f"Invalid or duplicate image name tag: {self._tag}, full image name: {self._full_name}"
        else:
            return f"Invalid or duplicate image name tag: {self._tag}"


class ProjectMismatchWithCanonical(ValueError):
    """
    Represent the project value does not match the canonical value when parsing the string representing the image in ImageRef.
    """

    def __init__(self, project: str, canonical: str) -> None:
        super().__init__(project, canonical)
        self._project = project
        self._canonical = canonical

    def __str__(self) -> str:
        return f'Project "{self._project}" mismatch with the image canonical: {self._canonical}'


class AliasResolutionFailed(ValueError):
    """
    Represents an alias resolution failure.
    The first argument of this exception should be the alias given by the user.
    """

    def __str__(self) -> str:
        return f"Failed to resolve alias: {self.args[0]}"


class InvalidIpAddressValue(ValueError):
    """
    Represents an invalid value for ip_address.
    """


class VolumeMountFailed(RuntimeError):
    """
    Represents a mount process failure.
    """


class VolumeUnmountFailed(RuntimeError):
    """
    Represents a umount process failure.
    """


class BackendError(web.HTTPError):
    """
    An RFC-7807 error class as a drop-in replacement of the original
    aiohttp.web.HTTPError subclasses.
    """

    error_type: str = "https://api.backend.ai/probs/general-error"
    error_title: str = "General Backend API Error."

    content_type: str
    extra_msg: Optional[str]

    body_dict: dict[str, Any]

    def __init__(self, extra_msg: str | None = None, extra_data: Optional[Any] = None, **kwargs):
        super().__init__(**kwargs)
        self.args = (self.status_code, self.reason, self.error_type)
        self.empty_body = False
        self.content_type = "application/problem+json"
        self.extra_msg = extra_msg
        self.extra_data = extra_data
        body = {
            "type": self.error_type,
            "title": self.error_title,
        }
        if extra_msg is not None:
            body["msg"] = extra_msg
        if extra_data is not None:
            body["data"] = extra_data
        self.body_dict = body
        self.body = json.dumps(body).encode()

    def __str__(self):
        lines = []
        if self.extra_msg:
            lines.append(f"{self.error_title} ({self.extra_msg})")
        else:
            lines.append(self.error_title)
        if self.extra_data:
            lines.append(" -> extra_data: " + repr(self.extra_data))
        return "\n".join(lines)

    def __repr__(self):
        lines = []
        if self.extra_msg:
            lines.append(
                f"<{type(self).__name__}({self.status}): {self.error_title} ({self.extra_msg})>"
            )
        else:
            lines.append(f"<{type(self).__name__}({self.status}): {self.error_title}>")
        if self.extra_data:
            lines.append(" -> extra_data: " + repr(self.extra_data))
        return "\n".join(lines)

    def __reduce__(self):
        return (
            type(self),
            (),  # empty the constructor args to make unpickler to use
            # only the exact current state in __dict__
            self.__dict__,
        )


class MalformedRequestBody(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Malformed request body."


class InvalidAPIParameters(BackendError, web.HTTPBadRequest):
    error_type = "https://api.backend.ai/probs/generic-bad-request"
    error_title = "Invalid or Missing API Parameters."


class MiddlewareParamParsingFailed(BackendError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Middleware parameter parsing failed."


class ParameterNotParsedError(BackendError, web.HTTPInternalServerError):
    error_type = "https://api.backend.ai/probs/internal-server-error"
    error_title = "Parameter Not Parsed Error"



import os
from pathlib import Path


def setns(fd: int):
    import ctypes
    import ctypes.util

    libc_path = ctypes.util.find_library("c")
    libc = ctypes.CDLL(libc_path)
    CLONE_NEWNET = 1 << 30
    libc.setns(fd, CLONE_NEWNET)


class NetworkNamespaceManager:
    def __init__(self, path: Path):
        self.self_ns = os.open("/proc/self/ns/net", os.O_RDONLY)
        self.new_ns = os.open(path, os.O_RDONLY)

    def __enter__(self):
        setns(self.new_ns)

    def __exit__(self, *exc_info_args):
        setns(self.self_ns)
        os.close(self.new_ns)
        os.close(self.self_ns)


def nsenter(path: Path) -> NetworkNamespaceManager:
    return NetworkNamespaceManager(path)



"""
Wrapper of msgpack-python with good defaults.
"""

import datetime
import enum
import os
import pickle
import uuid
from collections.abc import Mapping
from decimal import Decimal
from pathlib import PosixPath, PurePosixPath
from typing import Any, Callable, Optional, Protocol

import msgpack as _msgpack
import temporenc

from .types import BinarySize, ResourceSlot

__all__ = ("packb", "unpackb")


class ExtTypes(enum.IntEnum):
    # We can define up to 128 extension type identifiers.
    UUID = 1
    DATETIME = 2
    DECIMAL = 3
    POSIX_PATH = 4
    PURE_POSIX_PATH = 5
    ENUM = 6
    IMAGE_REF = 7
    RESOURCE_SLOT = 8
    BACKENDAI_BINARY_SIZE = 16


def _default(obj: object) -> _msgpack.ExtType:
    from .docker import ImageRef

    match obj:
        case tuple():
            return list(obj)
        case uuid.UUID():
            return _msgpack.ExtType(ExtTypes.UUID, obj.bytes)
        case datetime.datetime():
            return _msgpack.ExtType(ExtTypes.DATETIME, temporenc.packb(obj))
        case BinarySize():
            return _msgpack.ExtType(ExtTypes.BACKENDAI_BINARY_SIZE, pickle.dumps(obj, protocol=5))
        case Decimal():
            return _msgpack.ExtType(ExtTypes.DECIMAL, pickle.dumps(obj, protocol=5))
        case PosixPath():
            return _msgpack.ExtType(ExtTypes.POSIX_PATH, os.fsencode(obj))
        case PurePosixPath():
            return _msgpack.ExtType(ExtTypes.PURE_POSIX_PATH, os.fsencode(obj))
        case ResourceSlot():
            return _msgpack.ExtType(ExtTypes.RESOURCE_SLOT, pickle.dumps(obj, protocol=5))
        case enum.Enum():
            return _msgpack.ExtType(ExtTypes.ENUM, pickle.dumps(obj, protocol=5))
        case ImageRef():
            return _msgpack.ExtType(ExtTypes.IMAGE_REF, pickle.dumps(obj, protocol=5))
    raise TypeError(f"Unknown type: {obj!r} ({type(obj)})")


class ExtFunc(Protocol):
    def __call__(self, data: bytes) -> Any:
        pass


_DEFAULT_EXT_HOOK: Mapping[ExtTypes, ExtFunc] = {
    ExtTypes.UUID: lambda data: uuid.UUID(bytes=data),
    ExtTypes.DATETIME: lambda data: temporenc.unpackb(data).datetime(),
    ExtTypes.DECIMAL: lambda data: pickle.loads(data),
    ExtTypes.POSIX_PATH: lambda data: PosixPath(os.fsdecode(data)),
    ExtTypes.PURE_POSIX_PATH: lambda data: PurePosixPath(os.fsdecode(data)),
    ExtTypes.ENUM: lambda data: pickle.loads(data),
    ExtTypes.RESOURCE_SLOT: lambda data: pickle.loads(data),
    ExtTypes.BACKENDAI_BINARY_SIZE: lambda data: pickle.loads(data),
    ExtTypes.IMAGE_REF: lambda data: pickle.loads(data),
}


class _Deserializer:
    def __init__(self, mapping: Optional[Mapping[int, ExtFunc]] = None):
        self._ext_hook: dict[int, ExtFunc] = {}
        mapping = mapping or {}
        self._ext_hook = {**mapping}
        for ext_type, func in _DEFAULT_EXT_HOOK.items():
            if ext_type not in self._ext_hook:
                self._ext_hook[ext_type] = func

    @property
    def ext_hook(self) -> Callable[[int, bytes], Any]:
        def _hook_callable(code: int, data: bytes) -> Any:
            if code in self._ext_hook:
                return self._ext_hook[code](data)
            return _msgpack.ExtType(code, data)

        return _hook_callable


uuid_to_str: Mapping[int, ExtFunc] = {ExtTypes.UUID: lambda data: str(uuid.UUID(bytes=data))}

DEFAULT_PACK_OPTS = {
    "use_bin_type": True,  # bytes -> bin type (default for Python 3)
    "strict_types": True,  # do not serialize subclasses using superclasses
    "default": _default,
}

DEFAULT_UNPACK_OPTS = {
    "raw": False,  # assume str as UTF-8 (default for Python 3)
    "strict_map_key": False,  # allow using UUID as map keys
    "use_list": False,  # array -> tuple
    "ext_hook": _Deserializer().ext_hook,
}


def packb(data: Any, **kwargs) -> bytes:
    opts = {**DEFAULT_PACK_OPTS, **kwargs}
    ret = _msgpack.packb(data, **opts)
    if ret is None:
        return b""
    return ret


def unpackb(
    packed: bytes, ext_hook_mapping: Optional[Mapping[int, ExtFunc]] = None, **kwargs
) -> Any:
    opts = {**DEFAULT_UNPACK_OPTS, **kwargs}
    if ext_hook_mapping is not None:
        opts["ext_hook"] = _Deserializer(ext_hook_mapping).ext_hook
    return _msgpack.unpackb(packed, **opts)



from __future__ import annotations

import asyncio
import logging
import re
from abc import ABCMeta, abstractmethod
from typing import Any, ClassVar, Dict, Generic, Iterator, Mapping, Optional, Tuple, Type, TypeVar
from weakref import WeakSet

from ai.backend.common.asyncio import cancel_tasks
from ai.backend.logging import BraceStyleAdapter
from ai.backend.plugin.entrypoint import scan_entrypoints

from ..etcd import AsyncEtcd
from ..exception import ConfigurationError

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = (
    "AbstractPlugin",
    "BasePluginContext",
)


class AbstractPlugin(metaclass=ABCMeta):
    """
    The minimum generic plugin interface.
    """

    plugin_config: Mapping[str, Any]
    """
    ``plugin_config`` contains the plugin-specific configuration read from the etcd.
    """

    local_config: Mapping[str, Any]
    """
    ``local_config`` contains the configuration read from the disk TOML file of the current daemon.
    This configuration is only updated when restarting the daemon and thus plugins should assume
    that it's read-only and immutable during its lifetime.
    e.g., If the plugin is running with the manager, it's the validated content of manager.toml file.
    """

    config_watch_enabled: ClassVar[bool] = True
    """
    If set True (default), the hosting plugin context will watch and automatically update
    the etcd's plugin configuration changes via the ``update_plugin_config()`` method.
    """

    def __init__(self, plugin_config: Mapping[str, Any], local_config: Mapping[str, Any]) -> None:
        """
        Instantiate the plugin with the given initial configuration.
        """
        self.plugin_config = plugin_config
        self.local_config = local_config

    @abstractmethod
    async def init(self, context: Optional[Any] = None) -> None:
        """
        Initialize any resource used by the plugin.
        """
        pass

    @abstractmethod
    async def cleanup(self) -> None:
        """
        Clean up any resource used by the plugin upon server cleanup.
        """
        pass

    @abstractmethod
    async def update_plugin_config(self, plugin_config: Mapping[str, Any]) -> None:
        """
        Handle runtime configuration updates.
        The config parameter contains both the updated parts
        and unchanged parts of the configuration.

        The default implementation is just to replace the config property,
        but actual plugins may trigger other operations to reflect config changes
        and/or inspect the differences of configs before replacing the current config.
        """
        self.plugin_config = plugin_config


P = TypeVar("P", bound=AbstractPlugin)


class BasePluginContext(Generic[P]):
    """
    A minimal plugin manager which controls the lifecycles of the given plugins
    and watches & applies the configuration changes in etcd.

    The subclasses must redefine ``plugin_group``.
    """

    etcd: AsyncEtcd
    local_config: Mapping[str, Any]
    plugins: Dict[str, P]
    plugin_group: ClassVar[str] = "backendai_XXX_v10"
    allowlist: ClassVar[Optional[set[str]]] = None
    blocklist: ClassVar[Optional[set[str]]] = None

    _config_watchers: WeakSet[asyncio.Task]

    def __init__(self, etcd: AsyncEtcd, local_config: Mapping[str, Any]) -> None:
        self.etcd = etcd
        self.local_config = local_config
        self.plugins = {}
        self._config_watchers = WeakSet()
        if m := re.search(r"^backendai_(\w+)_v(\d+)$", self.plugin_group):
            self._group_key = m.group(1)
        else:
            raise TypeError(
                f"{type(self).__name__} has invalid plugin_group class attribute",
                self.plugin_group,
            )

    @classmethod
    def discover_plugins(
        cls,
        plugin_group: str,
        allowlist: Optional[set[str]] = None,
        blocklist: Optional[set[str]] = None,
    ) -> Iterator[Tuple[str, Type[P]]]:
        cls_allowlist = set() if cls.allowlist is None else cls.allowlist
        arg_allowlist = set() if allowlist is None else allowlist
        allowlist_enabled = allowlist is not None or cls.allowlist is not None
        cls_blocklist = set() if cls.blocklist is None else cls.blocklist
        arg_blocklist = set() if blocklist is None else blocklist
        for entrypoint in scan_entrypoints(
            plugin_group,
            allowlist=cls_allowlist | arg_allowlist if allowlist_enabled else None,
            blocklist=cls_blocklist | arg_blocklist,
        ):
            log.info("loading plugin (group:{}): {}", plugin_group, entrypoint.name)
            yield entrypoint.name, entrypoint.load()

    async def init(
        self,
        context: Optional[Any] = None,
        allowlist: Optional[set] = None,
        blocklist: Optional[set] = None,
    ) -> None:
        if allowlist is not None and blocklist is not None:
            if union := allowlist & blocklist:
                raise ConfigurationError({
                    "plugin.BasePluginContext": f"allowlist and blocklist has union value '{union}'"
                })
        scanned_plugins = self.discover_plugins(
            self.plugin_group,
            allowlist=allowlist,
            blocklist=blocklist,
        )
        for plugin_name, plugin_entry in scanned_plugins:
            plugin_config = await self.etcd.get_prefix(
                f"config/plugins/{self._group_key}/{plugin_name}/",
            )
            try:
                plugin_instance = plugin_entry(plugin_config, self.local_config)
                await plugin_instance.init(context=context)
            except Exception:
                log.exception("error during initialization of plugin: {}", plugin_name)
                continue
            else:
                self.plugins[plugin_name] = plugin_instance
            if plugin_instance.config_watch_enabled:
                await self.watch_config_changes(plugin_name)
        await asyncio.sleep(0)

    async def cleanup(self) -> None:
        await cancel_tasks(self._config_watchers)
        await asyncio.sleep(0)
        for plugin_instance in self.plugins.values():
            await plugin_instance.cleanup()

    async def _watcher(self, plugin_name: str) -> None:
        # As wait_timeout applies to the waiting for an internal async queue,
        # so short timeouts for polling the changes does not incur gRPC/network overheads.
        async for _ in self.etcd.watch_prefix(
            f"config/plugins/{self._group_key}/{plugin_name}",
            wait_timeout=0.2,
        ):
            new_config = await self.etcd.get_prefix(
                f"config/plugins/{self._group_key}/{plugin_name}/",
            )
            await self.plugins[plugin_name].update_plugin_config(new_config)

    async def watch_config_changes(self, plugin_name: str) -> None:
        wtask = asyncio.create_task(self._watcher(plugin_name))
        self._config_watchers.add(wtask)



from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Optional, Self, TypeVar, override

T = TypeVar("T")


@dataclass
class AbstractAgentResp(ABC):
    @abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError


@dataclass
class PurgeImageResp(AbstractAgentResp):
    image: str
    error: Optional[str] = None

    @override
    def as_dict(self) -> dict:
        return asdict(self)

    @classmethod
    def success(cls, image: str) -> Self:
        return cls(image)

    @classmethod
    def failure(cls, image: str, error: str) -> Self:
        return cls(image, error)


@dataclass
class PurgeImagesResp(AbstractAgentResp):
    responses: list[PurgeImageResp]

    @override
    def as_dict(self) -> dict:
        return asdict(self)



from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Any, Dict, Mapping, MutableMapping, Optional, Tuple, Union, cast

import humps
import tomli
import trafaret as t
from pydantic import (
    BaseModel,
    ConfigDict,
)

from . import validators as tx
from .etcd import AsyncEtcd, ConfigScopes
from .exception import ConfigurationError
from .types import RedisHelperConfig

__all__ = (
    "ConfigurationError",
    "etcd_config_iv",
    "redis_config_iv",
    "redis_helper_config_iv",
    "redis_helper_default_config",
    "vfolder_config_iv",
    "model_definition_iv",
    "read_from_file",
    "read_from_etcd",
    "override_key",
    "override_with_env",
    "check",
    "merge",
)


class BaseSchema(BaseModel):
    model_config = ConfigDict(
        validate_by_name=True,
        from_attributes=True,
        use_enum_values=True,
        extra="allow",
    )


etcd_config_iv = t.Dict({
    t.Key("etcd"): t.Dict({
        t.Key("namespace"): t.String,
        t.Key("addr", ("127.0.0.1", 2379)): tx.HostPortPair,
        t.Key("user", default=""): t.Null | t.String(allow_blank=True),
        t.Key("password", default=""): t.Null | t.String(allow_blank=True),
    }).allow_extra("*"),
}).allow_extra("*")

redis_helper_default_config: RedisHelperConfig = {
    "socket_timeout": 5.0,
    "socket_connect_timeout": 2.0,
    "reconnect_poll_timeout": 0.3,
}

redis_helper_config_iv = t.Dict({
    t.Key("socket_timeout", default=5.0): t.ToFloat,
    t.Key("socket_connect_timeout", default=2.0): t.ToFloat,
    t.Key("reconnect_poll_timeout", default=0.3): t.ToFloat,
}).allow_extra("*")

redis_default_config = {
    "addr": None,
    "sentinel": None,
    "service_name": None,
    "password": None,
    "redis_helper_config": redis_helper_default_config,
}

redis_config_iv = t.Dict({
    t.Key("addr", default=redis_default_config["addr"]): t.Null | tx.HostPortPair,
    t.Key(  # if present, addr is ignored and service_name becomes mandatory.
        "sentinel", default=redis_default_config["sentinel"]
    ): t.Null | tx.DelimiterSeperatedList(tx.HostPortPair),
    t.Key("service_name", default=redis_default_config["service_name"]): t.Null | t.String,
    t.Key("password", default=redis_default_config["password"]): t.Null | t.String,
    t.Key(
        "redis_helper_config",
        default=redis_helper_default_config,
    ): redis_helper_config_iv,
    t.Key("override_configs", default=None): t.Null
    | t.Mapping(
        t.String,
        t.Dict({
            t.Key("addr", default=redis_default_config["addr"]): t.Null | tx.HostPortPair,
            t.Key(  # if present, addr is ignored and service_name becomes mandatory.
                "sentinel", default=redis_default_config["sentinel"]
            ): t.Null | tx.DelimiterSeperatedList(tx.HostPortPair),
            t.Key("service_name", default=redis_default_config["service_name"]): t.Null | t.String,
            t.Key("password", default=redis_default_config["password"]): t.Null | t.String,
            t.Key(
                "redis_helper_config",
                default=redis_helper_default_config,
            ): redis_helper_config_iv,
        }).allow_extra("*"),
    ),
}).allow_extra("*")

vfolder_config_iv = t.Dict({
    tx.AliasedKey(["mount", "_mount"], default=None): t.Null | tx.Path(type="dir"),
    tx.AliasedKey(["fsprefix", "_fsprefix"], default=""): tx.Path(
        type="dir", resolve=False, relative_only=True, allow_nonexisting=True
    ),
}).allow_extra("*")

# Used in Etcd as a global config.
# If `scalingGroup.scheduler_opts` contains an `agent_selector_config`, it will override this.
agent_selector_globalconfig_iv = t.Dict({}).allow_extra("*")

# Used in `scalingGroup.scheduler_opts` as a per scaling_group config.
agent_selector_config_iv = t.Dict({}) | agent_selector_globalconfig_iv


model_definition_iv = t.Dict({
    t.Key("models"): t.List(
        t.Dict({
            t.Key("name"): t.String,
            t.Key("model_path"): t.String,
            t.Key("service", default=None): t.Null
            | t.Dict({
                # ai.backend.kernel.service.ServiceParser.start_service()
                # ai.backend.kernel.service_actions
                t.Key("pre_start_actions", default=[]): t.Null
                | t.List(
                    t.Dict({
                        t.Key("action"): t.String,
                        t.Key("args"): t.Dict().allow_extra("*"),
                    })
                ),
                t.Key("start_command"): t.String | t.List(t.String),
                t.Key("shell", default="/bin/bash"): t.String,  # used if start_command is a string
                t.Key("port"): t.ToInt[1:],
                t.Key("health_check", default=None): t.Null
                | t.Dict({
                    t.Key("interval", default=10): t.Null | t.ToFloat[0:],
                    t.Key("path"): t.String,
                    t.Key("max_retries", default=10): t.Null | t.ToInt[1:],
                    t.Key("max_wait_time", default=15): t.Null | t.ToFloat[0:],
                    t.Key("expected_status_code", default=200): t.Null | t.ToInt[100:],
                }),
            }),
            t.Key("metadata", default=None): t.Null
            | t.Dict({
                t.Key("author", default=None): t.Null | t.String(allow_blank=True),
                t.Key("title", default=None): t.Null | t.String(allow_blank=True),
                t.Key("version", default=None): t.Null | t.Int | t.String,
                tx.AliasedKey(["created", "created_at"], default=None): t.Null
                | t.String(allow_blank=True),
                tx.AliasedKey(["last_modified", "modified_at"], default=None): t.Null
                | t.String(allow_blank=True),
                t.Key("description", default=None): t.Null | t.String(allow_blank=True),
                t.Key("task", default=None): t.Null | t.String(allow_blank=True),
                t.Key("category", default=None): t.Null | t.String(allow_blank=True),
                t.Key("architecture", default=None): t.Null | t.String(allow_blank=True),
                t.Key("framework", default=None): t.Null | t.List(t.String),
                t.Key("label", default=None): t.Null | t.List(t.String),
                t.Key("license", default=None): t.Null | t.String(allow_blank=True),
                t.Key("min_resource", default=None): t.Null | t.Dict().allow_extra("*"),
            }).allow_extra("*"),
        })
    )
})


def find_config_file(daemon_name: str) -> Path:
    toml_path_from_env = os.environ.get("BACKEND_CONFIG_FILE", None)
    if not toml_path_from_env:
        toml_paths = [
            Path.cwd() / f"{daemon_name}.toml",
        ]
        if sys.platform.startswith("linux") or sys.platform.startswith("darwin"):
            parent_path = Path.cwd().parent
            while parent_path.is_relative_to(Path.home()):
                if (parent_path / "BUILD_ROOT").exists():
                    toml_paths.append(parent_path / f"{daemon_name}.toml")
                parent_path = parent_path.parent
            toml_paths += [
                Path.home() / ".config" / "backend.ai" / f"{daemon_name}.toml",
                Path(f"/etc/backend.ai/{daemon_name}.toml"),
            ]
        else:
            raise ConfigurationError({
                "read_from_file()": (
                    f"Unsupported platform for config path auto-discovery: {sys.platform}"
                ),
            })
    else:
        toml_paths = [Path(toml_path_from_env)]
    for _path in toml_paths:
        if _path.is_file():
            return _path
    else:
        searched_paths = ",".join(map(str, toml_paths))
        raise ConfigurationError({
            "find_config_file()": f"Could not read config from: {searched_paths}",
        })


def read_from_file(
    toml_path: Optional[Union[Path, str]], daemon_name: str
) -> Tuple[Dict[str, Any], Path]:
    config: Dict[str, Any]
    discovered_path: Path
    if toml_path is None:
        discovered_path = find_config_file(daemon_name)
    else:
        discovered_path = Path(toml_path)
    try:
        config = cast(Dict[str, Any], tomli.loads(discovered_path.read_text()))
    except IOError:
        raise ConfigurationError({
            "read_from_file()": f"Could not read config from: {discovered_path}",
        })
    else:
        return config, discovered_path


async def read_from_etcd(
    etcd_config: Mapping[str, Any], scope_prefix_map: Mapping[ConfigScopes, str]
) -> Optional[Dict[str, Any]]:
    etcd = AsyncEtcd(etcd_config["addr"], etcd_config["namespace"], scope_prefix_map)
    raw_value = await etcd.get("daemon/config")
    if raw_value is None:
        return None
    config: Dict[str, Any]
    config = cast(Dict[str, Any], tomli.loads(raw_value))
    return config


def override_key(table: MutableMapping[str, Any], key_path: Tuple[str, ...], value: Any):
    for k in key_path[:-1]:
        if k not in table:
            table[k] = {}
        table = table[k]
    table[key_path[-1]] = value


def override_with_env(table: MutableMapping[str, Any], key_path: Tuple[str, ...], env_key: str):
    val = os.environ.get(env_key, None)
    if val is None:
        return
    override_key(table, key_path, val)


def check(table: Any, iv: t.Trafaret):
    try:
        config = iv.check(table)
    except t.DataError as e:
        raise ConfigurationError(e.as_dict())
    else:
        return config


def merge(table: Mapping[str, Any], updates: Mapping[str, Any]) -> Mapping[str, Any]:
    result = {**table}
    for k, v in updates.items():
        if isinstance(v, Mapping):
            orig = result.get(k, {})
            assert isinstance(orig, Mapping)
            result[k] = merge(orig, v)
        else:
            result[k] = v
    return result


def set_if_not_set(table: MutableMapping[str, Any], key_path: Tuple[str, ...], value: Any) -> None:
    for k in key_path[:-1]:
        if k not in table:
            return
        table = table[k]
    if table.get(key_path[-1]) is None:
        table[key_path[-1]] = value


def config_key_to_snake_case(o: Any) -> Any:
    match o:
        case dict():
            return {humps.dekebabize(k): config_key_to_snake_case(v) for k, v in o.items()}
        case list() | tuple() | set():
            return [config_key_to_snake_case(i) for i in o]
        case _:
            return o



from __future__ import annotations

import asyncio
import base64
import ipaddress
import numbers
import random
import re
import sys
import uuid
from collections import OrderedDict
from datetime import timedelta
from itertools import chain
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import aiofiles
import yarl
from async_timeout import timeout

if TYPE_CHECKING:
    from decimal import Decimal

    from aiofiles.threadpool.text import AsyncTextIOWrapper

# It is a bad practice to keep all "miscellaneous" stuffs
# into the single "utils" module.
# Let's categorize them by purpose and domain, and keep
# refactoring to use the proper module names.

from .asyncio import (  # for legacy imports  # noqa
    AsyncBarrier,
    cancel_tasks,
    current_loop,
    run_through,
)
from .enum_extension import StringSetFlag  # for legacy imports  # noqa
from .files import AsyncFileWriter  # for legacy imports  # noqa
from .networking import curl, find_free_port  # for legacy imports  # noqa
from .types import BinarySize
from .exception import VolumeMountFailed, VolumeUnmountFailed
from .defs import DEFAULT_FILE_IO_TIMEOUT

KT = TypeVar("KT")
VT = TypeVar("VT")


def env_info() -> str:
    """
    Returns a string that contains the Python version and runtime path.
    """
    v = sys.version_info
    pyver = f"Python {v.major}.{v.minor}.{v.micro}"
    if v.releaselevel == "alpha":
        pyver += "a"
    if v.releaselevel == "beta":
        pyver += "b"
    if v.releaselevel == "candidate":
        pyver += "rc"
    if v.releaselevel != "final":
        pyver += str(v.serial)
    return f"{pyver} (env: {sys.prefix})"


def odict(*args: Tuple[KT, VT]) -> OrderedDict[KT, VT]:
    """
    A short-hand for the constructor of OrderedDict.
    :code:`odict(('a',1), ('b',2))` is equivalent to
    :code:`OrderedDict([('a',1), ('b',2)])`.
    """
    return OrderedDict(args)


def dict2kvlist(o: Mapping[KT, VT]) -> Iterable[Union[KT, VT]]:
    """
    Serializes a dict-like object into a generator of the flatten list of
    repeating key-value pairs.  It is useful when using HMSET method in Redis.

    Example:

       >>> list(dict2kvlist({'a': 1, 'b': 2}))
       ['a', 1, 'b', 2]
    """
    return chain.from_iterable((k, v) for k, v in o.items())


def generate_uuid() -> str:
    u = uuid.uuid4()
    # Strip the last two padding characters because u always has fixed length.
    return base64.urlsafe_b64encode(u.bytes)[:-2].decode("ascii")


def get_random_seq(length: float, num_points: int, min_distance: float) -> Iterator[float]:
    """
    Generate a random sequence of numbers within the range [0, length]
    with the given number of points and the minimum distance between the points.

    Note that X ( = the minimum distance d x the number of points N) must be equivalent to or smaller than
    the length L + d to guarantee the the minimum distance between the points.
    If X == L + d, the points are always equally spaced with d.

    :return: An iterator over the generated sequence
    """
    assert num_points * min_distance <= length + min_distance, (
        "There are too many points or it has a too large distance which cannot be fit into the"
        " given length."
    )
    extra = length - (num_points - 1) * min_distance
    ro = [random.uniform(0, 1) for _ in range(num_points + 1)]
    sum_ro = sum(ro)
    rn = [extra * r / sum_ro for r in ro[0:num_points]]
    spacing = [min_distance + rn[i] for i in range(num_points)]
    cumulative_sum = 0.0
    for s in spacing:
        cumulative_sum += s
        yield cumulative_sum - min_distance


def nmget(
    o: Mapping[str, Any],
    key_path: str,
    def_val: Optional[Any] = None,
    path_delimiter: str = ".",
    null_as_default: bool = True,
) -> Any:
    """
    A short-hand for retrieving a value from nested mappings
    ("nested-mapping-get"). At each level it checks if the given "path"
    component in the given key exists and return the default value whenever
    fails.

    Example:
    >>> o = {'a':{'b':1}, 'x': None}
    >>> nmget(o, 'a', 0)
    {'b': 1}
    >>> nmget(o, 'a.b', 0)
    1
    >>> nmget(o, 'a/b', 0, '/')
    1
    >>> nmget(o, 'a.c', 0)
    0
    >>> nmget(o, 'x', 0)
    0
    >>> nmget(o, 'x', 0, null_as_default=False)
    None
    """
    pieces = key_path.split(path_delimiter)
    while pieces:
        p = pieces.pop(0)
        if o is None or p not in o:
            return def_val
        o = o[p]
    if o is None and null_as_default:
        return def_val
    return o


def readable_size_to_bytes(expr: Any) -> BinarySize | Decimal:
    if isinstance(expr, numbers.Integral):
        return BinarySize(expr)
    return BinarySize.from_str(expr)


def str_to_timedelta(tstr: str) -> timedelta:
    """
    Convert humanized timedelta string into a Python timedelta object.

    Example:
    >>> str_to_timedelta('30min')
    datetime.timedelta(seconds=1800)
    >>> str_to_timedelta('1d1hr')
    datetime.timedelta(days=1, seconds=3600)
    >>> str_to_timedelta('2hours 15min')
    datetime.timedelta(seconds=8100)
    >>> str_to_timedelta('20sec')
    datetime.timedelta(seconds=20)
    >>> str_to_timedelta('300')
    datetime.timedelta(seconds=300)
    >>> str_to_timedelta('-1day')
    datetime.timedelta(days=-1)
    """
    _rx = re.compile(
        r"(?P<sign>[+|-])?\s*"
        r"((?P<days>\d+(\.\d+)?)(d|day|days))?\s*"
        r"((?P<hours>\d+(\.\d+)?)(h|hr|hrs|hour|hours))?\s*"
        r"((?P<minutes>\d+(\.\d+)?)(m|min|mins|minute|minutes))?\s*"
        r"((?P<seconds>\d+(\.\d+)?)(s|sec|secs|second|seconds))?$"
    )
    match = _rx.match(tstr)
    if not match:
        try:
            return timedelta(seconds=float(tstr))  # consider bare number string as seconds
        except TypeError:
            pass
        raise ValueError("Invalid time expression")
    groups = match.groupdict()
    sign = groups.pop("sign", None)
    if set(groups.values()) == {None}:
        raise ValueError("Invalid time expression")
    params = {n: -float(t) if sign == "-" else float(t) for n, t in groups.items() if t}
    return timedelta(**params)  # type: ignore


class FstabEntry:
    """
    Entry class represents a non-comment line on the `fstab` file.
    """

    def __init__(
        self, device: str, mountpoint: str, fstype: str, options: str | None, d=0, p=0
    ) -> None:
        self.device = device
        self.mountpoint = mountpoint
        self.fstype = fstype
        if not options:
            options = "defaults"
        self.options = options
        self.d = d
        self.p = p

    def __eq__(self, o):
        return str(self) == str(o)

    def __str__(self):
        return "{} {} {} {} {} {}".format(
            self.device, self.mountpoint, self.fstype, self.options, self.d, self.p
        )


class Fstab:
    """
    Reader/writer for fstab file.
    Takes aiofile pointer for async I/O. It should be writable if add/remove
    operations are needed.

    NOTE: This class references Jorge Niedbalski R.'s gist snippet.
          We have been converted it to be compatible with Python 3
          and to support async I/O.
          (https://gist.github.com/niedbalski/507e974ed2d54a87ad37)
    """

    def __init__(self, fp: AsyncTextIOWrapper) -> None:
        self._fp = fp

    def _hydrate_entry(self, line: str) -> FstabEntry:
        return FstabEntry(*[x for x in line.strip("\n").split(" ") if x not in ("", None)])

    async def get_entries(self) -> AsyncIterator[FstabEntry]:
        await self._fp.seek(0)
        for line in await self._fp.readlines():
            try:
                line = line.strip()
                if not line.startswith("#"):
                    yield self._hydrate_entry(line)
            except TypeError:
                pass

    async def get_entry_by_attr(self, attr: str, value: Any) -> FstabEntry | None:
        async for entry in self.get_entries():
            e_attr = getattr(entry, attr)
            if e_attr == value:
                return entry
        return None

    async def add_entry(self, entry: FstabEntry) -> None:
        if await self.get_entry_by_attr("device", entry.device):
            return
        await self._fp.write(str(entry) + "\n")
        await self._fp.truncate()

    async def add(
        self, device: str, mountpoint: str, fstype: str, options: str | None = None, d=0, p=0
    ) -> None:
        return await self.add_entry(FstabEntry(device, mountpoint, fstype, options, d, p))

    async def remove_entry(self, entry: FstabEntry) -> bool:
        await self._fp.seek(0)
        lines = await self._fp.readlines()
        line_no: int | None = None
        for index, line in enumerate(lines):
            if not line.strip().startswith("#"):
                try:
                    if self._hydrate_entry(line) == entry:
                        line_no = index
                        break
                except TypeError:
                    pass
        else:
            return False
        assert line_no is not None
        del lines[line_no]
        await self._fp.seek(0)
        await self._fp.write("".join(lines))
        await self._fp.truncate()
        return True

    async def remove_by_mountpoint(self, mountpoint: str) -> bool:
        entry = await self.get_entry_by_attr("mountpoint", mountpoint)
        if entry:
            return await self.remove_entry(entry)
        return False


async def mount(
    mount_path: str,
    fs_location: str,
    fs_type: str = "nfs",
    cmd_options: str | None = None,
    edit_fstab: bool = False,
    fstab_path: str | None = None,
    mount_prefix: str | None = None,
) -> None:
    if mount_prefix is None:
        mount_prefix = "/"
    if fstab_path is None:
        fstab_path = "/etc/fstab"
    mountpoint = Path(mount_prefix) / mount_path
    mountpoint.mkdir(exist_ok=True)
    if cmd_options is not None:
        cmd = [
            "mount",
            "-t",
            fs_type,
            "-o",
            cmd_options,
            fs_location,
            str(mountpoint),
        ]
    else:
        cmd = ["mount", "-t", fs_type, fs_location, str(mountpoint)]
    proc = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    raw_out, raw_err = await proc.communicate()
    raw_out.decode("utf8")
    err = raw_err.decode("utf8")
    await proc.wait()
    if err:
        raise VolumeMountFailed(f"Failed to mount {fs_location} on {mountpoint}. (e: {err})")
    if edit_fstab:
        async with aiofiles.open(fstab_path, mode="r+") as fp:  # type: ignore
            fstab = Fstab(fp)
            await fstab.add(
                fs_location,
                str(mountpoint),
                fs_type,
                cmd_options,
            )


async def umount(
    mount_path: str,
    mount_prefix: str | None = None,
    edit_fstab: bool = False,
    fstab_path: str | None = None,
    rmdir_if_empty: bool = False,
    *,
    timeout_sec: float | None = DEFAULT_FILE_IO_TIMEOUT,
) -> bool:
    if mount_prefix is None:
        mount_prefix = "/"
    if fstab_path is None:
        fstab_path = "/etc/fstab"
    mountpoint = Path(mount_prefix) / mount_path
    assert Path(mount_prefix) != mountpoint
    if not mountpoint.is_mount():
        return False
    try:
        with timeout(timeout_sec):
            proc = await asyncio.create_subprocess_exec(
                *[
                    "umount",
                    str(mountpoint),
                ],
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            raw_out, raw_err = await proc.communicate()
            raw_out.decode("utf8")
            err = raw_err.decode("utf8")
            await proc.wait()
    except asyncio.TimeoutError:
        raise VolumeUnmountFailed(
            f"Failed to umount {mountpoint}. Raise timeout ({timeout_sec}sec). "
            "The process may be hanging in state D, which needs to be checked."
        )
    if err:
        raise VolumeUnmountFailed(f"Failed to umount {mountpoint}")
    if rmdir_if_empty:
        try:
            mountpoint.rmdir()  # delete directory if empty
        except OSError:
            pass
    if edit_fstab:
        async with aiofiles.open(fstab_path, mode="r+") as fp:  # type: ignore
            fstab = Fstab(fp)
            await fstab.remove_by_mountpoint(str(mountpoint))
    return True


def is_ip_address_format(str: str) -> bool:
    try:
        url = yarl.URL("//" + str)
        if url.host and ipaddress.ip_address(url.host):
            return True
        return False
    except ValueError:
        return False


def join_non_empty(*args: Optional[str], sep: str) -> str:
    """
    Joins non-empty strings from the given arguments using the specified separator.
    """
    filtered_args = [arg for arg in args if arg]
    return sep.join(filtered_args)


def b64encode(s: str) -> str:
    """
    base64 encoding method of graphql_relay.
    Use it in components where the graphql_relay package is unavailable.
    """
    b: bytes = s.encode("utf-8") if isinstance(s, str) else s
    return base64.b64encode(b).decode("ascii")



from pydantic import Field

from ...api_handlers import BaseRequestModel


class PurgeImagesReq(BaseRequestModel):
    images: list[str] = Field(
        description="List of image canonical names to be purged",
    )
    force: bool = Field(
        description="Remove the images even if it is being used by stopped containers or has other tags",
        default=False,
    )
    noprune: bool = Field(
        description="Don't delete untagged parent images",
        default=False,
    )



from typing import Optional

from pydantic import Field

from ...api_handlers import BaseFieldModel
from ...types import VolumeID


class VolumeMetaField(BaseFieldModel):
    volume_id: VolumeID = Field(description="Used to uniquely identify a volume for operations.")
    backend: str = Field(description="Specifies the storage backend to determine handling methods.")
    path: str = Field(description="Defines the volume's location for access and management.")
    fsprefix: Optional[str] = Field(
        description="Indicates the filesystem prefix for path resolution."
    )
    capabilities: list[str] = Field(
        description="Lists allowed operations like read or write access."
    )


class VFolderMetaField(BaseFieldModel):
    mount_path: str = Field(description="Defines where the vfolder is mounted for access.")
    file_count: int = Field(description="Tracks the number of files to monitor storage usage.")
    used_bytes: int = Field(description="Indicates the current storage usage for quota checks.")
    capacity_bytes: int = Field(description="Defines the maximum allowed storage capacity.")
    fs_used_bytes: int = Field(description="Includes metadata and overhead in filesystem usage.")



from dataclasses import dataclass

import pyroscope


@dataclass
class PyroscopeArgs:
    enabled: bool
    application_name: str
    server_address: str
    sample_rate: int


class Profiler:
    def __init__(self, pyroscope_args: PyroscopeArgs) -> None:
        if pyroscope_args.enabled:
            pyroscope.configure(
                application_name=pyroscope_args.application_name,
                server_address=pyroscope_args.server_address,
                sample_rate=pyroscope_args.sample_rate,
            )



import uuid
from typing import Optional

from pydantic import AliasChoices, Field

from ...api_handlers import BaseRequestModel
from ...typed_validators import VFolderName
from ...types import VFolderUsageMode
from .field import VFolderPermissionField


class VFolderCreateReq(BaseRequestModel):
    name: VFolderName = Field(description="Name of the vfolder")
    folder_host: Optional[str] = Field(default=None, alias="host")
    usage_mode: VFolderUsageMode = Field(default=VFolderUsageMode.GENERAL)
    permission: VFolderPermissionField = Field(default=VFolderPermissionField.READ_WRITE)
    unmanaged_path: Optional[str] = Field(default=None, alias="unmanagedPath")
    group_id: Optional[uuid.UUID] = Field(
        default=None,
        validation_alias=AliasChoices("group", "groupId"),
    )
    cloneable: bool = Field(default=False)


class RenameVFolderReq(BaseRequestModel):
    new_name: VFolderName = Field(description="Name of the vfolder")



from typing import Awaitable, Callable, Optional, TypeAlias

from aiohttp import web

from ai.backend.common.contexts.request_id import with_request_id

Handler: TypeAlias = Callable[
    [web.Request],
    Awaitable[web.StreamResponse],
]

REQUEST_ID_HEADER = "X-BackendAI-RequestID"


@web.middleware
async def request_id_middleware(request: web.Request, handler: Handler) -> web.StreamResponse:
    _handler = handler
    request_id: Optional[str] = request.headers.get(REQUEST_ID_HEADER, None)
    with with_request_id(request_id):
        resp = await _handler(request)
    return resp



from __future__ import annotations

import enum
from abc import ABCMeta, abstractmethod
from typing import Any, Mapping, Optional, Union

from . import AbstractPlugin, BasePluginContext

__all__ = (
    "AbstractStatReporterPlugin",
    "AbstractErrorReporterPlugin",
    "StatsPluginContext",
    "ErrorPluginContext",
    "INCREMENT",
    "GAUGE",
)


class StatMetricTypes(enum.Enum):
    INCREMENT = 0
    GAUGE = 1


INCREMENT = StatMetricTypes.INCREMENT
GAUGE = StatMetricTypes.GAUGE


class AbstractStatReporterPlugin(AbstractPlugin, metaclass=ABCMeta):
    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    @abstractmethod
    async def report_metric(
        self,
        metric_type: StatMetricTypes,
        metric_name: str,
        value: Optional[Union[float, int]] = None,
    ) -> None:
        pass


class AbstractErrorReporterPlugin(AbstractPlugin, metaclass=ABCMeta):
    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    @abstractmethod
    async def capture_exception(
        self,
        exc_instance: Optional[Exception] = None,
        context: Optional[Mapping[str, Any]] = None,
    ) -> None:
        pass

    @abstractmethod
    async def capture_message(self, message: str) -> None:
        pass


class StatsPluginContext(BasePluginContext[AbstractStatReporterPlugin]):
    plugin_group = "backendai_stats_monitor_v20"

    async def report_metric(
        self,
        metric_type: StatMetricTypes,
        metric_name: str,
        value: Optional[Union[float, int]] = None,
    ) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.report_metric(metric_type, metric_name, value)


class ErrorPluginContext(BasePluginContext[AbstractErrorReporterPlugin]):
    plugin_group = "backendai_error_monitor_v20"

    async def capture_exception(
        self,
        exc_instance: Optional[Exception] = None,
        context: Optional[Mapping[str, Any]] = None,
    ) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.capture_exception(exc_instance, context)

    async def capture_message(self, message: str) -> None:
        for plugin_instance in self.plugins.values():
            await plugin_instance.capture_message(message)



import re
from enum import StrEnum
from typing import Final

# Redis database IDs depending on purposes
REDIS_STATISTICS_DB: Final = 0
REDIS_RATE_LIMIT_DB: Final = 1
REDIS_LIVE_DB: Final = 2
REDIS_IMAGE_DB: Final = 3
REDIS_STREAM_DB: Final = 4
REDIS_STREAM_LOCK: Final = 5


class RedisRole(StrEnum):
    STATISTICS = "statistics"
    RATE_LIMIT = "rate_limit"
    LIVE = "live"
    IMAGE = "image"
    STREAM = "stream"
    STREAM_LOCK = "stream_lock"


DEFAULT_FILE_IO_TIMEOUT: Final = 10

_RESERVED_VFOLDER_PATTERNS = [r"^\.[a-z0-9]+rc$", r"^\.[a-z0-9]+_profile$"]
RESERVED_VFOLDERS = [
    ".terminfo",
    ".jupyter",
    ".tmux.conf",
    ".ssh",
    "/bin",
    "/boot",
    "/dev",
    "/etc",
    "/lib",
    "/lib64",
    "/media",
    "/mnt",
    "/opt",
    "/proc",
    "/root",
    "/run",
    "/sbin",
    "/srv",
    "/sys",
    "/tmp",
    "/usr",
    "/var",
    "/home",
]
RESERVED_VFOLDER_PATTERNS = [re.compile(x) for x in _RESERVED_VFOLDER_PATTERNS]
API_VFOLDER_LENGTH_LIMIT: Final[int] = 64
MODEL_VFOLDER_LENGTH_LIMIT: Final[int] = 128

DEFAULT_VFOLDER_PERMISSION_MODE: Final[int] = 0o755
VFOLDER_GROUP_PERMISSION_MODE: Final[int] = 0o775

DEFAULT_DOMAIN_NAME: Final[str] = "default"

NOOP_STORAGE_VOLUME_NAME: Final[str] = "noop"
NOOP_STORAGE_BACKEND_TYPE: Final[str] = "noop"



from typing import NewType

from callosum.auth import (
    AbstractClientAuthenticator,
    AbstractServerAuthenticator,
    AuthResult,
    Credential,
    Identity,
)

PublicKey = NewType("PublicKey", bytes)
SecretKey = NewType("SecretKey", bytes)


class ManagerAuthHandler(AbstractClientAuthenticator):
    def __init__(
        self,
        domain: str,
        agent_public_key: PublicKey,
        manager_public_key: PublicKey,
        manager_secret_key: SecretKey,
    ) -> None:
        self.domain = domain
        self._agent_public_key = agent_public_key
        self._manager_public_key = manager_public_key
        self._manager_secret_key = manager_secret_key

    async def server_public_key(self) -> bytes:
        return self._agent_public_key

    async def client_public_key(self) -> bytes:
        return self._manager_public_key

    async def client_identity(self) -> Identity:
        assert self._manager_secret_key is not None
        return Identity(self.domain, self._manager_secret_key)


class AgentAuthHandler(AbstractServerAuthenticator):
    def __init__(
        self,
        domain: str,
        manager_public_key: PublicKey,
        agent_public_key: PublicKey,
        agent_secret_key: SecretKey,
    ) -> None:
        self.domain = domain
        self._manager_public_key = manager_public_key
        self._agent_public_key = agent_public_key
        self._agent_secret_key = agent_secret_key

    async def server_public_key(self) -> bytes:
        return self._agent_public_key

    async def server_identity(self) -> Identity:
        assert self._agent_secret_key is not None
        return Identity(self.domain, self._agent_secret_key)

    async def check_client(self, creds: Credential) -> AuthResult:
        if creds.domain == self.domain and creds.public_key == self._manager_public_key:
            return AuthResult(success=True, user_id="manager")
        return AuthResult(success=False)



from typing import Optional

from pydantic import Field

from ...api_handlers import BaseRequestModel
from ...types import QuotaConfig, VFolderID


class QuotaScopeReq(BaseRequestModel):
    options: Optional[QuotaConfig] = Field(
        default=None,
        description="The options for the quota scope.",
    )


class GetVFolderMetaReq(BaseRequestModel):
    subpath: str = Field(
        description="The subpath of the virtual folder.",
    )


class CloneVFolderReq(BaseRequestModel):
    dst_vfolder_id: VFolderID = Field(
        description="The destination virtual folder ID.",
        alias="dst_vfid",
    )
